from typing import (
    Callable,
    ClassVar,
    Union,
    Optional,
    cast,
    TYPE_CHECKING,
    Iterable,
    Any,
    Literal,
    TypedDict,
    Unpack,
    NotRequired,
    Awaitable,
    overload,
)
from itertools import chain, takewhile, zip_longest
from functools import lru_cache, cache
from inspect import isclass

from pydantic import BaseModel

from .reference import FirestoreCollection
from .document import GenericModel
from .snapshot import FirestoreSnapshot
from .utils import to_flatten


if TYPE_CHECKING:
    from google.api_core.retry import Retry
    from google.cloud.firestore import (
        Transaction,
        AsyncTransaction,
        Client,
        AsyncClient,
        WriteOption,
        Query,
        AsyncQuery,
        WriteBatch,
        AsyncWriteBatch,
        DocumentReference,
        AsyncDocumentReference,
        CollectionReference,
        AsyncCollectionReference,
    )
    from google.cloud.firestore_v1.types import WriteResult
    from google.cloud.firestore_v1.document import Timestamp
    from google.cloud.firestore_v1.bulk_writer import BulkWriter


Source = Union[
    "Client",
    "AsyncClient",
    "Transaction",
    "AsyncTransaction",
    "WriteBatch",
    "AsyncWriteBatch",
    "BulkWriter",
]


class PydanticParams(TypedDict):
    include: NotRequired[Iterable[str] | None]
    exclude: NotRequired[Iterable[str] | None]
    by_alias: NotRequired[bool]
    exclude_unset: NotRequired[bool]
    exclude_defaults: NotRequired[bool]
    exclude_none: NotRequired[bool]
    round_trip: NotRequired[bool]
    warnings: NotRequired[bool | Literal["none", "warn", "error"]]
    fallback: NotRequired[Callable[[Any], Any] | None]
    serialize_as_any: NotRequired[bool]
    context: NotRequired[Any | None]


class FirestoreBaseParams(TypedDict):
    retry: NotRequired[Optional["Retry"]]
    timeout: NotRequired[float | None]


class FirestoreCreateParams(FirestoreBaseParams):
    attempts: NotRequired[int | None]


class FirestoreReadParams(FirestoreBaseParams):
    field_paths: NotRequired[Iterable[str] | None]


class FirestoreUpdateParams(FirestoreCreateParams):
    option: NotRequired[Optional["WriteOption"]]


class FirestoreSetParams(FirestoreCreateParams):
    merge: NotRequired[bool]


class FirestoreWriteParams(FirestoreSetParams, FirestoreUpdateParams):
    pass


class CreateParams(FirestoreCreateParams, PydanticParams):
    pass


class UpdateParams(FirestoreUpdateParams, PydanticParams):
    pass


class SetParams(FirestoreSetParams, PydanticParams):
    pass


class FirestoreDict(TypedDict, total=False):
    id_field: str
    path_fields: tuple[str, ...]
    location: Union[str, tuple[str, ...]]


@cache
def default_param():
    from google.api_core.gapic_v1.method import DEFAULT

    return DEFAULT


class FirestoreHandler:
    def create(
        self,
        reference: "DocumentReference",
        data: dict,
        **kwargs: Unpack[FirestoreCreateParams],
    ) -> Optional["WriteResult"]:
        return reference.create(
            data,
            timeout=kwargs.get("timeout"),
            retry=kwargs.get("retry", default_param()),
        )

    def set(
        self,
        reference: "DocumentReference",
        data: dict,
        **kwargs: Unpack[FirestoreSetParams],
    ) -> Optional["WriteResult"]:
        return reference.set(
            data,
            merge=kwargs.get("merge", False),
            timeout=kwargs.get("timeout"),
            retry=kwargs.get("retry", default_param()),
        )

    def update(
        self,
        reference: "DocumentReference",
        data: dict,
        **kwargs: Unpack[FirestoreUpdateParams],
    ) -> Optional["WriteResult"]:
        return reference.update(
            data,
            timeout=kwargs.get("timeout"),
            retry=kwargs.get("retry", default_param()),
            option=kwargs.get("option"),
        )

    def delete(
        self,
        reference: "DocumentReference",
        **kwargs: Unpack[FirestoreUpdateParams],
    ) -> Optional["Timestamp"]:
        return reference.delete(
            timeout=kwargs.get("timeout"),
            retry=kwargs.get("retry", default_param()),
            option=kwargs.get("option"),
        )

    def write(
        self,
        method: Literal["create", "set", "update", "delete"],
        reference: "DocumentReference",
        data: dict | None = None,
        **kwargs: Unpack[FirestoreWriteParams],
    ) -> "Union[None, WriteResult, Timestamp]":
        return getattr(self, method)(reference, data=data, **kwargs)


class AsyncFirestoreHandler:
    async def create(
        self,
        reference: "AsyncDocumentReference",
        data: dict,
        **kwargs: Unpack[FirestoreCreateParams],
    ) -> Optional["WriteResult"]:
        return await reference.create(
            data,
            timeout=kwargs.get("timeout"),
            retry=kwargs.get("retry", default_param()),
        )

    async def set(
        self,
        reference: "AsyncDocumentReference",
        data: dict,
        **kwargs: Unpack[FirestoreSetParams],
    ) -> Optional["WriteResult"]:
        return await reference.set(
            data,
            merge=kwargs.get("merge", False),
            timeout=kwargs.get("timeout"),
            retry=kwargs.get("retry", default_param()),
        )

    async def update(
        self,
        reference: "AsyncDocumentReference",
        data: dict,
        **kwargs: Unpack[FirestoreUpdateParams],
    ) -> Optional["WriteResult"]:
        return await reference.update(
            data,
            timeout=kwargs.get("timeout"),
            retry=kwargs.get("retry", default_param()),
            option=kwargs.get("option"),
        )

    async def delete(
        self,
        reference: "AsyncDocumentReference",
        **kwargs: Unpack[FirestoreUpdateParams],
    ) -> Optional["Timestamp"]:
        return await reference.delete(
            timeout=kwargs.get("timeout"),
            retry=kwargs.get("retry", default_param()),
            option=kwargs.get("option"),
        )

    async def write(
        self,
        method: Literal["create", "set", "update", "delete"],
        reference: "AsyncDocumentReference",
        data: dict | None = None,
        **kwargs: Unpack[FirestoreWriteParams],
    ) -> "Union[None, WriteResult, Timestamp]":
        return await getattr(self, method)(reference, data=data, **kwargs)


class TransactionHandler(FirestoreHandler):
    def __init__(
        self,
        writer: "Union[Transaction, AsyncTransaction, WriteBatch, AsyncWriteBatch]",
    ):
        self.writer = writer

    def create(self, reference, data, **kwargs):
        self.writer.create(reference, data)

    def set(self, reference, data, **kwargs):
        self.writer.set(
            reference,
            data,
            merge=kwargs.get("merge", False),
        )

    def update(self, reference, data, **kwargs):
        self.writer.update(
            reference,
            data,
            option=kwargs.get("option"),
        )

    def delete(self, reference, **kwargs):
        self.writer.delete(reference, option=kwargs.get("option"))


class BulkHandler(FirestoreHandler):
    def __init__(self, writer: "BulkWriter"):
        self.writer = writer

    def create(self, reference, data, **kwargs):
        self.writer.create(reference, data, attempts=kwargs.get("attempts", 0))

    def set(self, reference, data, **kwargs):
        self.writer.set(
            reference,
            data,
            merge=kwargs.get("merge", False),
            attempts=kwargs.get("attempts", 0),
        )

    def update(self, reference, data, **kwargs):
        self.writer.update(
            reference,
            data,
            option=kwargs.get("option"),
            attempts=kwargs.get("attempts", 0),
        )

    def delete(self, reference, **kwargs):
        self.writer.delete(
            reference, option=kwargs.get("option"), attempts=kwargs.get("attempts", 0)
        )


fallback_handler = FirestoreHandler()
async_fallback_handler = AsyncFirestoreHandler()


@lru_cache
def source_to_handler(
    source: Source,
) -> tuple[
    "Union[Client, AsyncClient]", Union[FirestoreHandler, AsyncFirestoreHandler]
]:
    from google.cloud.firestore import (
        Transaction,
        WriteBatch,
        Client,
        AsyncClient,
        AsyncWriteBatch,
        AsyncTransaction,
    )
    from google.cloud.firestore_v1.bulk_writer import BulkWriter

    if isinstance(source, (Transaction, AsyncTransaction, WriteBatch, AsyncWriteBatch)):
        return source._client, TransactionHandler(source)
    elif isinstance(source, BulkWriter):
        return source._client, BulkHandler(source)
    elif isinstance(source, Client):
        return source, fallback_handler
    elif isinstance(source, AsyncClient):
        return source, async_fallback_handler
    else:
        raise ValueError("Invalid firestore source")


@lru_cache
def source_to_tuple(
    source: "Union[Client, Transaction, WriteBatch, BulkWriter]",
) -> "tuple[Client, Optional[Transaction]]":
    from google.cloud.firestore import Transaction, Client

    if isinstance(source, Transaction):
        return source._client, source
    elif isinstance(source, Client):
        return source, None
    else:
        raise ValueError("Invalid firestore source")


@lru_cache
def async_source_to_tuple(
    source: "Union[Client, AsyncClient, AsyncTransaction, AsyncWriteBatch]",
) -> "tuple[AsyncClient, Optional[AsyncTransaction]]":
    from google.cloud.firestore import AsyncTransaction, AsyncClient

    if isinstance(source, AsyncTransaction):
        return source._client, source
    elif isinstance(source, AsyncClient):
        return source, None
    else:
        raise ValueError("Invalid firestore source")


class FirestoreModel(BaseModel):
    firestore_id_field: ClassVar[str | None] = None
    firestore_path_fields: ClassVar[tuple[str, ...]] = ()
    firestore_location: ClassVar[tuple[str, ...]] = ()
    firestore_config: ClassVar[FirestoreDict] = {}

    def __init_subclass__(cls, **kwargs):
        if (id_field := cls.firestore_config.get("id_field")) is not None:
            cls.firestore_id_field = str(id_field)
        if (location := cls.firestore_config.get("location")) is not None:
            cls.firestore_location = (
                (location,) if isinstance(location, str) else location
            )
        if (path_fields := cls.firestore_config.get("path_fields")) is not None:
            cls.firestore_path_fields = path_fields

        if cls.firestore_location != ():
            for value in cls.__dict__.values():
                if isclass(value) and issubclass(value, FirestoreModel):
                    value.firestore_location = cls.firestore_location
                    value.firestore_id_field = cls.firestore_id_field
                    value.firestore_path_fields = cls.firestore_path_fields

        super().__init_subclass__(**kwargs)

    @classmethod
    @lru_cache
    def firestore_collection(cls, *args: str):
        assert len(cls.firestore_location) > 0
        assert len(args) == len(cls.firestore_location) - 1, "Invalid args count"
        return FirestoreCollection(
            takewhile(
                lambda x: x is not None,
                chain.from_iterable(
                    zip_longest(cls.firestore_location, reversed(args))
                ),
            )
        )

    @classmethod
    def firestore_document(cls, id: str, *args: str):
        return cls.firestore_collection(*args).document(id)

    @overload
    @classmethod
    def document_reference(
        cls, client: "Client", id: str, *args: str
    ) -> "DocumentReference": ...
    @overload
    @classmethod
    def document_reference(
        cls, client: "AsyncClient", id: str, *args: str
    ) -> "AsyncDocumentReference": ...
    @classmethod
    def document_reference(
        cls, client: "Union[Client, AsyncClient]", id: str, *args: str
    ) -> "Union[DocumentReference, AsyncDocumentReference]":
        return cls.firestore_document(id, *args).to_firestore(client)

    @overload
    @classmethod
    def collection_reference(
        cls, client: "Client", *args: str
    ) -> "CollectionReference": ...
    @overload
    @classmethod
    def collection_reference(
        cls, client: "AsyncClient", *args: str
    ) -> "AsyncCollectionReference": ...
    @classmethod
    def collection_reference(
        cls, client: "Union[Client, AsyncClient]", *args: str
    ) -> "Union[CollectionReference, AsyncCollectionReference]":
        return cls.firestore_collection(*args).to_firestore(client)

    @classmethod
    def firestore_dump(
        cls, data: GenericModel, **kwargs: Unpack[PydanticParams]
    ) -> dict[str, Any]:
        return data.__pydantic_serializer__.to_python(
            data,
            mode="python",
            by_alias=kwargs.get("by_alias", True),
            exclude=kwargs.get("exclude"),
            include=kwargs.get("include"),
            exclude_unset=kwargs.get("exclude_unset", False),
            exclude_defaults=kwargs.get("exclude_defaults", False),
            exclude_none=kwargs.get("exclude_none", False),
            warnings=kwargs.get("warnings", True),
            round_trip=kwargs.get("round_trip", False),
            fallback=kwargs.get("fallback"),
            serialize_as_any=kwargs.get("serialize_as_any", False),
            context=kwargs.get("context"),
        )

    @classmethod
    def firestore_write(
        cls,
        method: Literal["create", "set", "update", "delete"],
        source: Source,
        id: str,
        *args: str,
        data: dict[str, Any] | None = None,
        **kwargs: Unpack[FirestoreWriteParams],
    ):
        client, handler = source_to_handler(source)
        return handler.write(
            method, cls.document_reference(client, id, *args), data=data, **kwargs
        )

    @classmethod
    def firestore_create(
        cls,
        source: Source,
        data: GenericModel,
        id: str,
        *args: str,
        **kwargs: Unpack[CreateParams],
    ):
        return cls.firestore_write(
            "create",
            source,
            id,
            *args,
            data=cls.firestore_dump(data, **kwargs),
            **kwargs,
        )

    @classmethod
    def firestore_set(
        cls,
        source: Source,
        data: GenericModel,
        id: str,
        *args: str,
        **kwargs: Unpack[SetParams],
    ) -> Optional["WriteResult"]:
        return cls.firestore_write(
            "set",
            source,
            id,
            *args,
            data=cls.firestore_dump(data, **kwargs),
            **kwargs,
        )

    @classmethod
    def firestore_context(cls, **kwargs) -> dict[str]:
        return {
            "firestore_id_field": cls.firestore_id_field,
            "firestore_path_fields": cls.firestore_path_fields,
            **kwargs,
        }

    @overload
    @classmethod
    def firestore_read(
        cls: type[GenericModel],
        source: Union["Client", "Transaction"],
        id: str,
        *args: str,
        strict: bool = False,
        context: dict[str, Any] | None = None,
        **kwargs: Unpack[FirestoreReadParams],
    ) -> FirestoreSnapshot[GenericModel]:
        client, transaction = source_to_tuple(source)
        _cls = cast(type[FirestoreModel], cls)
        return FirestoreSnapshot[cls].from_firestore(
            _cls.document_reference(client, id, *args).get(
                transaction=transaction,
                field_paths=kwargs.get("field_paths"),
                timeout=kwargs.get("timeout"),
                retry=kwargs.get("retry"),
            ),
            strict=strict,
            context=_cls.firestore_context(**(context or {})),
        )

    @classmethod
    async def firestore_read_async(
        cls: type[GenericModel],
        source: Union["AsyncClient", "AsyncTransaction"],
        id: str,
        *args: str,
        strict: bool = False,
        context: dict[str, Any] | None = None,
        **kwargs: Unpack[FirestoreReadParams],
    ) -> FirestoreSnapshot[GenericModel]:
        client, transaction = async_source_to_tuple(source)
        _cls = cast(type[FirestoreModel], cls)
        return FirestoreSnapshot[cls].from_firestore(
            await _cls.document_reference(client, id, *args).get(
                transaction=transaction,
                field_paths=kwargs.get("field_paths"),
                timeout=kwargs.get("timeout"),
                retry=kwargs.get("retry"),
            ),
            strict=strict,
            context=_cls.firestore_context(**(context or {})),
        )

    @classmethod
    def firestore_update(
        cls,
        source: Source,
        data: GenericModel,
        id: str,
        *args: str,
        ignore_flatten: Iterable[str] | None = None,
        **kwargs: Unpack[UpdateParams],
    ) -> Optional["WriteResult"]:
        return cls.firestore_write(
            "update",
            source,
            id,
            *args,
            data=to_flatten(
                cls.firestore_dump(data, exclude_unset=True, **kwargs),
                exclude=ignore_flatten,
            ),
            **kwargs,
        )

    @classmethod
    def firestore_delete(
        cls,
        source: Source,
        id: str,
        *args: str,
        **kwargs: Unpack[FirestoreUpdateParams],
    ) -> Optional["Timestamp"]:
        return cls.firestore_write("delete", source, id, *args, **kwargs)

    @overload
    @classmethod
    def firestore_query(
        cls,
        source: "Union[Client, Transaction]",
        *args: str,
    ) -> "Query": ...
    @overload
    @classmethod
    def firestore_query(
        cls,
        source: "Union[AsyncClient, AsyncTransaction]",
        *args: str,
    ) -> "AsyncQuery": ...
    @classmethod
    def firestore_query(
        cls,
        source: "Union[Client, AsyncClient, Transaction, AsyncTransaction]",
        *args: str,
    ) -> "Union[Query, AsyncQuery]":
        client, _ = source_to_handler(source)
        return cls.collection_reference(client, *args)._query()

    def _firestore_path(self, *args: str) -> tuple[str, ...]:
        if (len_args := len(args)) == (len_loc := len(self.firestore_location)):
            id, *args = args
        elif len_args == (len_loc - 1):
            id = getattr(self, self.firestore_id_field)
        else:
            id, args = getattr(self, self.firestore_id_field), tuple(
                getattr(self, field_name) for field_name in self.firestore_path_fields
            )
        return (id, args)

    def create_to_firestore(
        self,
        source: Source,
        *args: str,
        context: dict[str, Any] | None = None,
        **kwargs: Unpack[CreateParams],
    ):
        id, args = self._firestore_path(*args)
        return self.firestore_create(source, self, id, *args, context=context, **kwargs)

    def set_to_firestore(
        self,
        source: Source,
        *args: str,
        context: dict[str, Any] | None = None,
        **kwargs: Unpack[SetParams],
    ):
        id, args = self._firestore_path(*args)
        return self.firestore_set(source, self, id, *args, context=context, **kwargs)

    def update_to_firestore(
        self,
        source: Source,
        *args: str,
        ignore_flatten: Iterable[str] | None = None,
        context: dict[str, Any] | None = None,
        **kwargs: Unpack[UpdateParams],
    ):
        id, args = self._firestore_path(*args)
        return self.firestore_update(
            source,
            self,
            id,
            *args,
            ignore_flatten=ignore_flatten,
            context=context,
            **kwargs,
        )

    def delete_to_firestore(
        self,
        source: Source,
        *args: str,
        **kwargs: Unpack[FirestoreUpdateParams],
    ):
        id, args = self._firestore_path(*args)
        return self.firestore_delete(source, id, *args, **kwargs)



from typing import Union, TYPE_CHECKING, Any, Iterable
from uuid import UUID
from decimal import Decimal
from functools import singledispatch
from collections.abc import MutableMapping
from ipaddress import IPv4Address, IPv6Address

from pydantic import PlainSerializer, AnyUrl
from pydantic_core import Url


if TYPE_CHECKING:
    from google.cloud.firestore_v1.field_path import FieldPath


@singledispatch
def to_firestore(value):
    if hasattr(value, "to_firestore"):
        return value.to_firestore()
    else:
        return value


@to_firestore.register
def _(value: Decimal) -> Union[int, float]:
    if value.as_integer_ratio()[1] == 1:
        return int(value)
    else:
        return float(value)


@to_firestore.register
def _(value: Url) -> str:
    return str(value)


@to_firestore.register
def _(value: AnyUrl) -> str:
    return str(value)


@to_firestore.register
def _(value: UUID) -> str:
    return str(value)


@to_firestore.register
def _(value: IPv4Address) -> str:
    return str(value)


@to_firestore.register
def _(value: IPv6Address) -> str:
    return str(value)


def _to_flatten(
    obj: dict[str, Any],
    *,
    exclude: set["FieldPath"],
    parent: "FieldPath",
) -> dict["FieldPath", Any]:
    items: list[tuple["FieldPath", Any]] = []

    for key, value in obj.items():
        field = parent + key
        if field not in exclude and isinstance(value, MutableMapping):
            items.extend(_to_flatten(value, parent=field, exclude=exclude).items())
        else:
            items.append((field, value))

    return dict(items)


def to_flatten(
    obj: dict[str, Any],
    *,
    exclude: Iterable[str] | None = None,
) -> dict[str, Any]:
    from google.cloud.firestore_v1.field_path import FieldPath

    if exclude is None:
        exclude = set()
    else:
        exclude = {FieldPath.from_string(path) for path in exclude}

    output = _to_flatten(obj, exclude=exclude, parent=FieldPath())

    return {field.to_api_repr(): value for field, value in output.items()}


firestore_serializer = PlainSerializer(to_firestore, return_type=object)



from typing import Any, Generic, TYPE_CHECKING, Annotated, Optional
from datetime import datetime

from pydantic import BaseModel, Field, ConfigDict, AliasChoices, AliasPath

from .reference import collection_validator, FirestoreCollection, FirestoreDocument
from .document import GenericModel, DocumentModel, document_validator

if TYPE_CHECKING:
    from google.cloud.firestore import DocumentSnapshot


def fix_snapshot(obj: "DocumentSnapshot") -> "DocumentSnapshot":
    for key in ("create_time", "update_time", "read_time"):
        value = obj.__dict__.get(key, None)

        if (
            value is not None
            and not isinstance(value, datetime)
            and hasattr(value, "ToDatetime")
        ):
            obj.__dict__[key] = value.ToDatetime()

    return obj


class FirestoreSnapshot(BaseModel, Generic[GenericModel]):
    id: str = Field(
        validation_alias=AliasChoices(AliasPath("_reference", "_path", -1), "id")
    )
    path: Annotated[FirestoreCollection, collection_validator] = Field(
        validation_alias=AliasChoices(AliasPath("_reference", "_path"), "path")
    )

    read_time: datetime

    create_time: datetime | None = Field(None, exclude=True)
    update_time: datetime | None = Field(None, exclude=True)

    document: Annotated[Optional[DocumentModel[GenericModel]], document_validator] = (
        Field(None, validation_alias=AliasChoices("_data", "document"))
    )

    model_config = ConfigDict(frozen=True)

    @property
    def reference(self) -> FirestoreDocument:
        return self.path.document(self.id)

    @property
    def exists(self) -> bool:
        return self.document is not None

    @property
    def data(self):
        return self.document.data if self.document is not None else None

    @classmethod
    def from_firestore(
        cls,
        snapshot: "DocumentSnapshot",
        strict: bool = False,
        context: dict[str, Any] | None = None,
    ) -> "FirestoreSnapshot[GenericModel]":
        ctx_fn = getattr(cls.__pydantic_generic_metadata__["args"][0], "firestore_context", lambda **kwargs: kwargs)
        return cls.__pydantic_validator__.validate_python(
            fix_snapshot(snapshot), strict=strict, from_attributes=True, context=ctx_fn(**(context or {}))
        )



from typing import Generic, TypeVar, Any
from datetime import datetime

from pydantic import (
    BaseModel,
    Field,
    ConfigDict,
    AliasChoices,
    ValidationInfo,
    WrapValidator,
    ValidatorFunctionWrapHandler,
)


GenericModel = TypeVar("GenericModel")


class DocumentModel(BaseModel, Generic[GenericModel]):
    create_time: datetime
    update_time: datetime

    data: GenericModel = Field(validation_alias=AliasChoices("_data", "data"))

    model_config = ConfigDict(frozen=True, from_attributes=False)


def _validate_document(
    value: Any, handler: ValidatorFunctionWrapHandler, info: ValidationInfo
):
    create_time = info.data.get("create_time")
    update_time = info.data.get("update_time")

    is_snapshot = create_time is not None and update_time is not None

    if is_snapshot and isinstance(value, dict):
        if isinstance(context := info.context, dict):
            if (id_field := context.get("firestore_id_field")) and isinstance(id_field, str):
                value.setdefault(id_field, info.data.get("id"))
            if (path_fields := context.get("firestore_path_fields")) and isinstance(path_fields, tuple):
                path = info.data.get("path")
                for (path_field, path_item) in zip(path_fields, info.data.get("path")):
                    value.setdefault(path_field, path_item)

        value = {
            "create_time": create_time,
            "update_time": update_time,
            "data": value,
        }

    return handler(value)


document_validator = WrapValidator(_validate_document)



from typing import TYPE_CHECKING, Any, overload, Union

from pydantic import (
    RootModel,
    ConfigDict,
    model_validator,
    Field,
    WrapValidator,
    ValidatorFunctionWrapHandler,
)

from .path import FirestorePath

if TYPE_CHECKING:
    from google.cloud.firestore import (
        DocumentReference,
        Client,
        CollectionReference,
        AsyncDocumentReference,
        AsyncCollectionReference,
        AsyncClient,
    )


class ReferenceModel(RootModel):
    root: FirestorePath = Field(kw_only=False)

    model_config = ConfigDict(frozen=True)

    def __iter__(self):
        return self.root.__iter__()

    def __str__(self) -> str:
        return str(self.root)

    @property
    def parent(self):
        parent = self.root.parent
        return self.__class__(parent) if parent is not None else None


class FirestoreDocument(ReferenceModel):
    @model_validator(mode="after")
    def check_path(self):
        assert self.root.is_document, f"Path is not a document: {self.root}"
        return self

    def collection(self, id: str) -> "FirestoreCollection":
        return FirestoreCollection(root=self.root / id)

    @property
    def parent(self):
        parent = self.root.parent
        return FirestoreCollection(root=parent) if parent is not None else None

    @classmethod
    def from_firestore(cls, value: "DocumentReference") -> "FirestoreDocument":
        return cls.model_validate(value._path)

    @overload
    def to_firestore(self, client: "AsyncClient") -> "AsyncDocumentReference": ...
    @overload
    def to_firestore(self, client: "Client") -> "DocumentReference": ...
    def to_firestore(
        self, client: "Union[Client, AsyncClient]"
    ) -> "Union[DocumentReference, AsyncDocumentReference]":
        return client.document(*self.root)


class FirestoreCollection(ReferenceModel):
    @model_validator(mode="after")
    def check_path(self):
        assert self.root.is_collection, f"Path is not a collection: {self.root}"
        return self

    def document(self, id: str) -> "FirestoreDocument":
        return FirestoreDocument(root=self.root / id)

    @property
    def parent(self):
        parent = self.root.parent
        return FirestoreDocument(root=parent) if parent is not None else None

    @classmethod
    def from_firestore(cls, value: "CollectionReference") -> "FirestoreCollection":
        return cls.model_validate(value._path)

    @overload
    def to_firestore(self, client: "AsyncClient") -> "AsyncCollectionReference": ...
    @overload
    def to_firestore(self, client: "Client") -> "CollectionReference": ...
    def to_firestore(
        self, client: "Union[Client, AsyncClient]"
    ) -> "Union[CollectionReference, AsyncCollectionReference]":
        return client.collection(*self.root)


def _validate_collection(value: Any, handler: ValidatorFunctionWrapHandler):
    if isinstance(value, FirestoreCollection):
        return value
    elif isinstance(value, tuple):
        path = FirestorePath(root=value)
        return FirestoreCollection(root=path.parent if path.is_document else path)
    else:
        return handler(value)


collection_validator = WrapValidator(_validate_collection)



from typing import Any, Self

from pydantic import (
    Field,
    RootModel,
    ConfigDict,
    model_serializer,
    model_validator,
    ValidationInfo,
)


class FirestorePath(RootModel):
    root: tuple[str, ...] = Field(kw_only=False)

    model_config = ConfigDict(frozen=True)

    @model_serializer(when_used="json")
    def to_json(self) -> str:
        return str(self)

    @model_validator(mode="before")
    def validate_path(value: Any, info: ValidationInfo):
        if isinstance(value, str):
            value = tuple(value.split("/"))

        return value

    @property
    def id(self) -> str:
        return self.root[-1]

    @property
    def root_id(self) -> str:
        return self.root[0]

    @property
    def parent_id(self) -> str | None:
        return self.root[-2]

    @property
    def is_root(self) -> bool:
        return len(self.root) == 1

    @property
    def is_document(self) -> bool:
        return len(self.root) % 2 == 0

    @property
    def is_collection(self) -> bool:
        return len(self.root) % 2 == 1

    @property
    def parent(self):
        return self.__class__(self.root[:-1]) if len(self.root) > 1 else None

    def __iter__(self):
        return self.root.__iter__()

    def __str__(self) -> str:
        return "/".join(self.root)

    def __truediv__(self, other) -> Self:
        return self.__class__((*self.root, str(other)))



from .path import FirestorePath
from .reference import FirestoreCollection, FirestoreDocument
from .snapshot import FirestoreSnapshot
from .model import FirestoreModel, FirestoreDict
from .transforms import *


__version__ = "0.5.0"



from pytest import fixture, mark

from google.cloud.firestore import DocumentReference, CollectionReference, Client

from pydantic_firestore.reference import FirestoreDocument, FirestoreCollection


@fixture
def client() -> Client:
    return Client()


DOC_EXAMPLES = [
    ("col", "0"),
    ("col", "0", "sub_col", "00"),
]

COL_EXAMPLES = [
    ("col",),
    ("col", "0", "sub_col"),
]


@mark.parametrize("path", DOC_EXAMPLES)
def test_from_document(path: tuple[str, ...]):
    doc_ref = DocumentReference(*path)
    fs_doc = FirestoreDocument.from_firestore(doc_ref)

    assert fs_doc.model_dump() == path


@mark.parametrize("path", DOC_EXAMPLES)
def test_to_document(path: tuple[str, ...], client: Client):
    fs_doc = FirestoreDocument(root=path)

    assert fs_doc.to_firestore(client)._path == path


@mark.parametrize("path", COL_EXAMPLES)
def test_from_collection(path: tuple[str, ...]):
    col_ref = CollectionReference(*path)
    fs_col = FirestoreCollection.from_firestore(col_ref)

    assert fs_col.model_dump() == path


@mark.parametrize("path", COL_EXAMPLES)
def test_to_collection(path: tuple[str, ...], client: Client):
    fs_col = FirestoreCollection(root=path)

    assert fs_col.to_firestore(client)._path == path


@mark.xfail
@mark.parametrize("path", COL_EXAMPLES)
def test_document_from_col_path(path: tuple[str, ...]):
    FirestoreDocument(root=path)


@mark.xfail
@mark.parametrize("path", DOC_EXAMPLES)
def test_collection_from_doc_path(path: tuple[str, ...]):
    FirestoreCollection(root=path)



from typing import Annotated, Literal, Union, Any, Callable, TypeVar, Tuple, TYPE_CHECKING
from enum import StrEnum, auto, global_enum
from functools import cache

from pydantic import WrapSerializer, ValidationInfo

from .utils import firestore_serializer


__all__ = [
    "Sentinel",
    "FIRESTORE_TIMESTAMP",
    "FIRESTORE_DELETE",
    "FIRESTORE_REMOVE",
    "FIRESTORE_UNION",
    "FIRESTORE_INCREMENT",
    "FIRESTORE_MAXIMUM",
    "FIRESTORE_MINIMUM",
    "FirestoreArray",
    "FirestoreNumeric",
    "FirestoreMinMax",
    "FirestoreTimestamp",
    "FirestoreDelete",
    "FirestoreSentinel",
    "sentinel_serializer",
]


@global_enum
class Sentinel(StrEnum):
    FIRESTORE_TIMESTAMP = auto()
    FIRESTORE_DELETE = auto()

    FIRESTORE_REMOVE = auto()
    FIRESTORE_UNION = auto()

    FIRESTORE_INCREMENT = auto()
    FIRESTORE_MAXIMUM = auto()
    FIRESTORE_MINIMUM = auto()

    @classmethod
    @cache
    def firestore_mapping(cls):
        from google.cloud.firestore_v1.transforms import (
            SERVER_TIMESTAMP,
            DELETE_FIELD,
            ArrayRemove,
            ArrayUnion,
            Increment,
            Maximum,
            Minimum,
        )

        return {
            cls.FIRESTORE_TIMESTAMP: lambda _: SERVER_TIMESTAMP,
            cls.FIRESTORE_DELETE: lambda _: DELETE_FIELD,
            cls.FIRESTORE_REMOVE: ArrayRemove,
            cls.FIRESTORE_UNION: ArrayUnion,
            cls.FIRESTORE_INCREMENT: Increment,
            cls.FIRESTORE_MAXIMUM: Maximum,
            cls.FIRESTORE_MINIMUM: Minimum,
        }

    def to_firestore(self, *, value=None, **kwargs) -> object:
        return self.__class__.firestore_mapping()[self](value)

    def __call__[T](self, value: T, /) -> "FirestoreSentinel[T]":
        return (self, value)


if TYPE_CHECKING:
    FIRESTORE_TIMESTAMP = Sentinel.FIRESTORE_TIMESTAMP
    FIRESTORE_DELETE = Sentinel.FIRESTORE_DELETE

    FIRESTORE_REMOVE = Sentinel.FIRESTORE_REMOVE
    FIRESTORE_UNION = Sentinel.FIRESTORE_UNION

    FIRESTORE_INCREMENT = Sentinel.FIRESTORE_INCREMENT
    FIRESTORE_MAXIMUM = Sentinel.FIRESTORE_MAXIMUM
    FIRESTORE_MINIMUM = Sentinel.FIRESTORE_MINIMUM

FirestoreArray = Literal[Sentinel.FIRESTORE_REMOVE, Sentinel.FIRESTORE_UNION]
FirestoreNumeric = Literal[
    Sentinel.FIRESTORE_INCREMENT, Sentinel.FIRESTORE_MAXIMUM, Sentinel.FIRESTORE_MINIMUM
]
FirestoreMinMax = Literal[Sentinel.FIRESTORE_MAXIMUM, Sentinel.FIRESTORE_MINIMUM]

FirestoreTimestamp = Annotated[
    Literal[Sentinel.FIRESTORE_TIMESTAMP], firestore_serializer
]
FirestoreDelete = Annotated[Literal[Sentinel.FIRESTORE_DELETE], firestore_serializer]


def _serialize_sentinel(value: Any, handler: Callable, info: ValidationInfo):
    if (
        isinstance(value, tuple)
        and len(value) == 2
        and (isinstance(value[0], Sentinel) or value[0] in Sentinel)
    ):
        sentinel, value = value
        return Sentinel(sentinel).to_firestore(value=handler(value))
    else:
        return handler(value)


sentinel_serializer = WrapSerializer(_serialize_sentinel, return_type=object)

T = TypeVar("T")
FirestoreSentinel = Annotated[Union[T, Tuple[Sentinel, T]], sentinel_serializer]
